/* 
   Teller, Copyright (C) 1996, 1997 James Bailie
   ==================================================================

   Teller is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Teller is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
   License for more details.

   Although Teller is licensed under the Free Software Foundation's GNU
   General Public License, Teller is not produced by, nor is it endorsed
   by the Free Software Foundation. The Free Software Foundation is
   not responsible for developing, distributing, or supporting Teller
   in any way. Anyone may place software they own the copyright to,
   under the GNU General Public License.

   The GNU General Public License is included in the News Peruser
   distribution archive in a file called COPYING. If you do not have a
   copy of the license, you can download one from
   ftp://prep.ai.mit.edu, or you can write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   =====================================================================
*/

#define BUFFER_SIZE 2048

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<ctype.h>
#include<netinet/in.h>
#include<netdb.h>
#include<unistd.h>


const char *name = "teller";
const char *home_directory;

int news_socket;  

FILE *news_server, *spool_file, *temp_file, *posted_file, *output;

char line_buffer[ BUFFER_SIZE ], copy_buffer[ BUFFER_SIZE ], 
  spool_path_and_name[ BUFFER_SIZE ];

typedef enum { stream_closed, stream_open } stream_state;

struct stream_record 
{
  stream_state news_socket_fd;
  stream_state news_server_stream;
  stream_state spool_file_stream;
  stream_state posted_file_stream;
  stream_state temp_file_stream;
    stream_state output_stream;
} open_streams;


void close_files( void )
{
  if ( open_streams.temp_file_stream )
    fclose( temp_file );

  if ( open_streams.spool_file_stream )
    fclose( spool_file );

  if ( open_streams.news_server_stream )
    fclose( news_server );

  if ( open_streams.news_socket_fd )
    close( news_socket );

  if ( open_streams.posted_file_stream )
    {
      fflush( posted_file );
      fclose( posted_file );
    }

  if ( open_streams.output_stream )
  {
      fflush( output );
      fclose( output );
  }

  return;
}

void graceful_exit( int signal_number )
{

/*
   Linux resets a signal's handler to its default once it has been raised,
   so we shut off the two signal's we handle to prevent interruption.
*/

  signal( signal_number, SIG_IGN );

  close_files();
  exit( signal_number );
}

void unrecoverable_error( char *server, char* message )
{
  fprintf( output, "teller: %s responded with a code\n", server );
  fprintf( output, "teller: I don't know how to handle:\n" );
  fprintf( output, "teller: %s", message );
  graceful_exit( SIGINT );
}


int main( int argc, char **argv[] )
{
  enum message_id_found { yes, no } found;

  char *line_pointer;

  int test;

  struct protoent *protocol;
  struct hostent *host;
  struct servent *service;
  struct sockaddr_in socket_address;


/*
   Establish signal handler.
*/

  signal( SIGINT, graceful_exit );
  signal( SIGHUP, graceful_exit );

/*
   Keep track of the state of streams.
*/

  open_streams.news_socket_fd = stream_closed;
  open_streams.news_server_stream = stream_closed;
  open_streams.spool_file_stream = stream_closed;
  open_streams.posted_file_stream = stream_closed;
  open_streams.temp_file_stream = stream_closed;
  open_streams.output_stream = stream_closed;

      /*
        Figure out if we've been execvp'ed from Peruser
      */

  if ( strstr( ( char *)argv[ 0 ], "cteller" ))
  {
      sprintf( line_buffer, "%s/.peruser_spool/client_pipe", getenv( "HOME" ));
      if (( output = fopen( line_buffer, "w" )) == NULL )
      {
          perror( name );
          graceful_exit( EXIT_FAILURE );
      }
      
      open_streams.output_stream = stream_open;
  }
  else
      output = stdout;
  
/*
   The NNTP server's name is passed as a command line argument.
*/
  
  if ( argc == 1 )
    {
      fprintf( output, "teller: You must supply the NNTP server's name\n" );
      fprintf( output, "teller: as a command line argument to Teller.\n" );
      graceful_exit( SIGINT );
    }

  if ( argc > 2 )
    {
      if ( !strncmp( ( char *)argv[ 2 ], "-a", 2 ))
	if ( argc != 5 )
	  {
	    fprintf( output, "teller: You must supply your NNTP server user\n" );
	    fprintf( output, "teller: name and password, respectively, after\n" );
	    fprintf( output, "teller: the -a option.\n" );
	    graceful_exit( SIGINT );
	  }
	else
	  {
	    fprintf( news_server, "authinfo user %s\r\n",
		     ( char *)argv[ 3 ] );
	    fprintf( output, "teller: authinfo user %s\n",
		     ( char *)argv[ 3 ] );
            fflush( output );

	    fgets( line_buffer, BUFFER_SIZE, news_server ); 
	    memcpy( copy_buffer, line_buffer, BUFFER_SIZE );
	    sprintf( line_buffer, "server: %s", copy_buffer );
	    fprintf( output, line_buffer );
            fflush( output );

	    if ( atoi( strtok( copy_buffer, " " )) != 381 )
	      unrecoverable_error( ( char *)argv[ 1 ], copy_buffer );

	    fprintf( news_server, "authinfo pass %s\r\n", 
		     ( char *)argv[ 4 ] );
	    fprintf( output, "teller: authinfo pass <password>\n" );
            fflush( output );

	    fgets( line_buffer, BUFFER_SIZE, news_server );
	    memcpy( copy_buffer, line_buffer, BUFFER_SIZE );
	    sprintf( line_buffer, "server: %s", copy_buffer );
	    fprintf( output, line_buffer );
            fflush( output );

	    if ( atoi( strtok( copy_buffer, " " )) != 281 )
	      unrecoverable_error( ( char *)argv[ 1 ], copy_buffer );
	  }
    }

/*
   Open up the FOLLOW-UPS spool file.
*/

  sprintf( spool_path_and_name, "%s/.peruser_spool/FOLLOW-UPS", 
	   home_directory = getenv( "HOME" ));
  if (( spool_file = fopen( spool_path_and_name, "r" )) == NULL )
    {
      if ( errno != 2 )
	{
	  perror( name );
	  graceful_exit( SIGINT );
	}
      else
	{
	  fprintf( output, "teller: No articles in queue!\n" );
	  graceful_exit( EXIT_FAILURE );
	}
    }

  open_streams.spool_file_stream = stream_open;

  test = fgetc( spool_file );

  if ( feof( spool_file ) )
    {
      fprintf( output, "teller: No articles in queue!\n" );
      graceful_exit( EXIT_FAILURE );
    }

  ungetc( test, spool_file );

/*
   Lookup the protocol number assigned to TCP. 
*/

  if ( !( protocol = getprotobyname( "tcp" )))
    {
      fprintf( output, "teller: Couldn't find a protocol number for \"tcp\"\n" );
      fprintf( output, "teller: in /etc/protocols.\n" );
      graceful_exit( SIGINT );
    }

/*
   Lookup the service number assigned to NNTP.
*/

  if ( !( service = getservbyname( "nntp", "tcp" )))
    {
      fprintf( output, "teller: Couldn't find a service number for \"nntp\"\n" );
      fprintf( output, "teller: in /etc/services.\n" );
      graceful_exit( SIGINT );
    }

/*
   Lookup the address of the NNTP server named on the command line.
*/

  if ( !( host = gethostbyname( ( char *)argv[ 1 ] )))
     {
       herror( name );
       graceful_exit( SIGINT );
     }

/*
   Create a socket for communication with the NNTP server.
*/

  if (( news_socket = socket( AF_INET, SOCK_STREAM, protocol->p_proto ))
      == -1 )
    {
      perror( name );
      graceful_exit( SIGINT );
    }

/*
   Zero and initialize the socket_address structure.
*/

  memset( &socket_address, 0, sizeof socket_address );
  socket_address.sin_family = AF_INET;
  memmove( (char *)&socket_address.sin_addr, host->h_addr, host->h_length );
  socket_address.sin_port = service->s_port;

/*
   Connect to the NNTP server through the socket.
*/

  if ( connect( news_socket, ( struct sockaddr *)&socket_address, 
	        sizeof socket_address ) == -1 )
    {
      perror( name );
      graceful_exit( SIGINT );
    }
  open_streams.news_socket_fd = stream_open;

/*
   Create a stream and associate it with the socket.
*/

  if ( !( news_server = fdopen( news_socket, "r+" )))
    {
      perror( name );
      graceful_exit( SIGINT );
    }
  open_streams.news_server_stream = stream_open;

/*
   Get the welcome message from the news server.
*/

  fgets( line_buffer, BUFFER_SIZE, news_server );
  fprintf( output, line_buffer );
  fflush( output );

/* 
   A three-digit response code starts all NNTP server replies. The
   three digits are represented as three characters of ASCII,
   separated from any message text by a single space.

   There are two acceptable response codes after a connection has been
   made to a NNTP server: "200" or "201". The "2" indicates a command
   succeeded, the zero the command that succeeded, in this case the
   connect. The last digit indicates either posting allowed: "0", or
   posting not allowed: "1".
*/

  if ( strncmp( line_buffer, "200", 3 ))
      {
	unrecoverable_error( ( char *)argv[ 1 ], line_buffer );
      }

/*
  Open the POSTED folder.
*/

  sprintf( line_buffer, "%s/.peruser_spool/POSTED",
	   home_directory );
  if (( posted_file = fopen( line_buffer, "a" )) == NULL )
    {
      perror( name );
      graceful_exit( EXIT_FAILURE );
    }
  open_streams.posted_file_stream = stream_open;


/*
   This is the main loop.   
*/

for( ; ; )
  {

/* 
   Temporary file to hold article until post confirmed.
*/

    sprintf( line_buffer, "%s/.peruser_spool/.temp", home_directory );
    if (( temp_file = fopen( line_buffer, "w+" )) == NULL )
      {
	perror( name );
	graceful_exit( EXIT_FAILURE );
      }
    open_streams.temp_file_stream = stream_open;

/*
  If the spool is empty, quit.
*/

    line_buffer[ 0 ] = fgetc( spool_file );
    if ( line_buffer[ 0 ] == EOF )
      break;
    ungetc( line_buffer[ 0 ], spool_file );

/*
   Ask the server to accept post. 
*/

    fputs( "post\r\n", news_server );
    fprintf( output, "teller: post\n" );
    fflush( output );
    fgets( line_buffer, BUFFER_SIZE, news_server );
    fprintf( output, "server: %s", line_buffer );
    fflush( output );
    if ( strncmp( line_buffer, "340", 3 ))
      unrecoverable_error( ( char *)argv[ 1 ], line_buffer );
    
/*
   Send article
*/

    found = no;
    for( ; ;)
      {
	fgets( line_buffer, BUFFER_SIZE, spool_file );
	fputs( line_buffer, temp_file );
	if ( !strncmp( line_buffer, ".\r\n", 3 ))
	  {
	    fputs( line_buffer, news_server );
	    fprintf( output,
                     "teller: Message sent. Awaiting confirmation of post...\n" );
            fflush( output );
	    break;
	  }
	if ( !strncmp( line_buffer, ".", 1 ))
	  {
	    strcpy( copy_buffer, line_buffer );
	    strcpy( line_buffer, "." );
	    strcat( line_buffer, copy_buffer );
	  }
	if ( found == no && ( !strncmp( line_buffer, "Message-ID: ", 12 ) ||
                              !strncmp( line_buffer, "Message-Id: ", 12 )))
	  {
	    line_pointer = strstr( line_buffer, " " );
	    line_pointer++;
	    fprintf( output,
                     "teller: sending message: %s", line_pointer );
            fflush( output );
            found = yes;
	  }
	fputs( line_buffer, news_server );
      }

/*
   Get server's response. 240 means post succeeded, 241 means post failed.
*/

    fgets( line_buffer, BUFFER_SIZE, news_server );
    fprintf( output, "server: %s\n", line_buffer );
    fflush( stdout );
    
    if ( strncmp( line_buffer, "240", 3 ))
      unrecoverable_error( ( char *)argv[ 1 ], line_buffer );

    rewind( temp_file );
    for( ; ; )
      {
	fgets( line_buffer, BUFFER_SIZE, temp_file );
	if ( feof( temp_file ))
	  break;

	fputs( line_buffer, posted_file );
      }
    fclose( temp_file );

  } /* end of main loop */

  fputs( "quit\r\n", news_server );
  fprintf( output, "teller: quit\n" );
  fflush( output );
  fgets( line_buffer, BUFFER_SIZE, news_server );
  fprintf( output, "server: %s", line_buffer );
  fflush( output );

  close_files();
  truncate( spool_path_and_name, 0 );
  sprintf( line_buffer, "%s./peruser_spool/.temp", home_directory );
  remove( line_buffer );

  fprintf( output, "\nteller: I have told the news.\n" );
  fflush( output );
  fclose( output );
  return( 0 );
}
